/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Generic cron-jobs. Currently contains only HELO
 * exchanges.
 * @author Christian Grothoff
 * @file server/heloexchange.c
 **/

#include "config.h"
#include "server/heloexchange.h"
#include "util/ipcheck.h"
#include "knownhosts.h"
#include <unistd.h>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include "statistics.h"
#include "server/pingpong.h"

#define TCP_HTTP_PORT 80
#define HTTP_URL "http://"
#define GET_COMMAND "GET %s HTTP/1.0\r\n\r\n"

/* ************* internal Methods **************** */

/**
 * Tell everybody we are there...
 **/
static void broadcastHELO(void * unused);
 
/**
 * Forward HELOs from all known hosts to all known hosts.
 **/
static void forwardHELO(void * unused);

/* ******************** CODE ********************* */

/**
 * Catch sigpipe on send, and do nothing
 **/
#ifndef LINUX
static void catcher(int sig) {
  MESSAGE("signal %d caught\n", sig);
  signal(sig, catcher);
}
#endif

/**
 * initialize a few cron jobs. Must be called after
 * initcron (!).
 **/
void initHeloExchange() {
  addCronJob(&broadcastHELO,
	     30,60*5,
	     NULL); /* every five minutes */
  addCronJob(&downloadHostlist,
	     15,0,
	     NULL); /* after 15s */
  if (YES == isHELOEnabled())
    addCronJob(&forwardHELO,
	       60, 7 *60,
	       NULL); /* seven minutes: exchange */
  else
    DEBUG("HELO forwarding disabled!\n");
}

/**
 * Stops a few cron jobs that exchange HELOs.
 **/
void doneHeloExchange() {
  delCronJob(&broadcastHELO,
	     60*5,
	     NULL); 
  delCronJob(&downloadHostlist,
	     0,
	     NULL); 
  if (YES == isHELOEnabled())
    delCronJob(&forwardHELO,
	       7 *60,
	       NULL); /* seven minutes: exchange */
}


/**
 * We have received a HELO. Verify (signature, integrity,
 * ping-pong) and store identity if ok.
 **/
void receivedHELO(HELO_Body * body) {
  HELO_Body * copy;
  HostAddress address;
  PublicKey pubKey;
  
  if (YES == isBlacklisted(body->sender.senderIP)) {
    INFO("Received HELO from blacklisted host. Dropping.\n");
    return; /* we don't like this IP (blacklisted) */  
  }
  if (SYSERR == verifyHELO(body)) {
    WARNING("HELO message received invalid. Dropping.\n");
    return; /* message invalid */  
  }
  if ( (OK == identity2Key(&body->senderIdentity,
			   &pubKey)) &&
       (OK == getConnectionInformation(&body->senderIdentity,
				       &address) ) ) {
    if ( (0 == memcmp(&address, &body->sender, sizeof(HostAddress))) &&
	 (0 == memcmp(&pubKey, &body->publicKey, sizeof(PublicKey))) ) {
      /* ok, we've seen this one exactly like this before (at most the
	 TTL has changed); thus we can 'trust' it without playing
	 ping-pong */
      bindAddress(body);
      return;
    }
  }
  copy = xmalloc(sizeof(HELO_Body),
		 "malloc for pingAction");
  memcpy(copy, body,
	 sizeof(HELO_Body));
  pingAction(&body->sender,
	     &body->senderIdentity,
	     (void (*)(void*))&bindAddress, 
	     copy);
}

/**
 * Download hostlist from the web.
 **/
static void downloadHostlistHelper(char * url) {
  unsigned short port;
  char * hostname;
  char * filename;
  int curpos;
  struct hostent *ip_info;
  struct sockaddr_in soaddr;
  int sock;
  int ret;
  char * command;
  HELO_Body body;
  time_t start;
  char c;

  port = TCP_HTTP_PORT;

  INFO("Trying download from %s\n",
       url);

  curpos = 0;
  if (0 != strncmp(HTTP_URL, url, strlen(HTTP_URL)) ) {
    WARNING("WARNING: invalid URL %s (must begin with %s)\n", url, HTTP_URL);
    DEBUG("CRON: exit downloadHostlist (error: not http)\n");
    return;
  }
  curpos = strlen(HTTP_URL);
  hostname = &url[curpos];
  while ( (curpos < strlen(url)) &&
	  (url[curpos] != '/') )
    curpos++;
  if (curpos == strlen(url))
    filename = strdup("/");
  else {
    filename = strdup(&url[curpos]);
    url[curpos] = '\0'; /* terminator for hostname */
  }
  ip_info = gethostbyname(hostname);
  if (ip_info == NULL) {
    WARNING("WARNING: could not download hostlist, host %s unknown\n", 
	    hostname);
    free(filename);
    free(url);
    ERROR("CRON: exit downloadHostlist (error: can not resolve hostname)\n");
    return;
  }
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    ERROR("ERROR: could not open socket for hostlist download %s\n",
	  strerror(errno));
    free(filename);
    free(url);
    ERROR("CRON: exit downloadHostlist (error: can not open socket)\n");
    return;
  }
  soaddr.sin_family = AF_INET;
  soaddr.sin_addr.s_addr = ((struct in_addr*)(ip_info->h_addr))->s_addr;
  soaddr.sin_port=htons(TCP_HTTP_PORT);
  if (connect (sock, (struct sockaddr*)&soaddr, sizeof(soaddr)) < 0) {
    print("WARNING: failed so send HTTP request to host %s (%d - %d) - %s\n",
	  hostname,
	  curpos, sock,
	  strerror(errno));
    free(filename);
    free(url);
    ERROR("CRON: exit downloadHostlist (error: can not connect)\n");
    return;
  }
  
  command = malloc(strlen(filename) + strlen(GET_COMMAND));
  sprintf(command, GET_COMMAND, filename);
  free(filename);
  curpos = strlen(command)+1;
#ifndef LINUX
  if ( SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    if (SIG_ERR == signal(SIGPIPE, catcher))
      fprintf(stderr, "WARNING: could not install handler for SIGPIPE!\n"\
                      "Attempting to continue anyway.");
  }
  curpos = send(sock, command, curpos, 0);
#else
  curpos = send(sock, command, curpos, MSG_NOSIGNAL);
#endif
  if (curpos != strlen(command)+1) {
    ERROR("WARNING: failed so send HTTP request %s to host %s (%d - %d) - %s\n",
	  command, hostname,
	  curpos, sock,
	  strerror(errno));
    free(url);
    free(command);
    ERROR("CRON: exit downloadHostlist (error)\n");
    return;
  }
  INFO("Receiving hostlist from %s ...\n",
       url);
   free(url);
  free(command);
  time(&start);

  /* we first have to read out the http_response*/
  /* it ends with four line delimiters: "\r\n\r\n" */
  curpos=0;
  while (curpos<4) {
    if (start + 30 < time(NULL))
      break; /* exit after 30s */
    ret = recv(sock,&c,sizeof(c), MSG_DONTWAIT);
    if (ret == 0)
      break;
    if (ret < 0) {
      if (errno != EAGAIN)
 	break;
      else
 	ret = 0;
    }
    if ((c=='\r') || (c=='\n')) 
      curpos += ret;
    else 
      curpos=0;    
  }
  
  if (curpos<4) { /* we have not found it */
    WARNING("CRON: exit downloadHostlist (error: no http response read)\n");
    return;
  }

  while (1) {
    if (start + 30 < time(NULL))
      break; /* exit after 30s */
    curpos = 0;
    while (curpos < sizeof(HELO_Body)) {
      if (start + 30 < time(NULL))
	break; /* exit after 30s */
      ret = recv(sock, 
		 &((char*)&body)[curpos],
		 sizeof(HELO_Body)-curpos,
		 MSG_DONTWAIT);
      if (ret == 0)
	break;
      if (ret < 0) {
	if (errno != EAGAIN)
	  break;
	else
	  ret = 0;
      }
      curpos += ret;
    }
    if (curpos != sizeof(HELO_Body))
      break;
    INFO(".");
    receivedHELO(&body);
  }
  close(sock);  
  INFO("\nCRON: exit downloadHostlist (success, %d seconds before timeout)\n",
       start + 30 - time(NULL));
}


/**
 * Download hostlist from the web.
 **/
void downloadHostlist(void * unused) {
  char * url;
  int i;
  
  DEBUG("CRON: enter downloadHostlist\n");
  url = getHostlistURL();
  if (url == NULL) {
    DEBUG("CRON: exit downloadHostlist (error: URL not specified)\n");
    return;
  }
  i = strlen(url);
  while (i > 0) {
    i--;
    if (url[i] == ' ') {
      downloadHostlistHelper(strdup(&url[i+1]));
      url[i] = '\0';
    }
  }
  downloadHostlistHelper(strdup(&url[0]));
  xfree(url, "temporary url copy (strdup)");  
}


typedef struct {
  /* the CRC of the message */
  int crc;
  /* the HELO message */
  HELO_Message m;
  /* send the HELO in 1 out of n cases */
  int n;
} SendData;

static void broadcastHelper(HostIdentity * hi,
			    SendData * sd) {
  if (randomi(sd->n) == 0) {
    GNUNET_STATISTICS.udp_out_counts[GNET_PROTO_HELO]++;
    sendToHost(NULL,
	       hi,
	       sizeof(HELO_Message),
	       &sd->m.header,
	       PLAINTEXT_FLAG,
	       sd->crc);
  }
}

/**
 * Tell a couple of random hosts on the currentKnownHost list 
 * that we are exist...
 **/
static void broadcastHELO(void * unused) {
  int j;
  SendData sd;

  DEBUG("CRON: enter broadcastHELO\n");
  j = forEachHost(NULL, 
		  0, 
		  NULL); /* just count */
  if (j <= 1) {
    printf("HELO: broadcast pointless: no hosts known\n");
    return; /* no point in trying... */
  }
  sd.n = j/((int) log(j+1)+1); /* send the HELO to log(known host) hosts */
  createHELO(&sd.m);
  bindAddress(&sd.m.body);
  sd.crc = crc32N(&sd.m, 
		  sizeof(HELO_Message));

  forEachHost((void (*)(HostIdentity*, void *))&broadcastHelper,
	      0, /* ignore blacklisting */
	      &sd);
  DEBUG("CRON: exit broadcastHELO\n");
}

/**
 * Forward HELOs from all known hosts to all connected hosts.
 **/
static void forwardHELOHelper(HostIdentity * identity,
			      int * count) {
  HELO_Message fi;
  FileName fn;
  time_t now;
  int ran;
  HexName hexName;

  DEBUG("FORWARDING HELOs\n");
  time(&now);
  fi.header.size = htons(sizeof(HELO_Message));
  fi.header.requestType = htons(GNET_PROTO_HELO); 
  fn = xmalloc(strlen((char*)getHostsDirectory())+sizeof(HexName)+1,
	       "forwardHELOHelper: filename");

  hash2hex(&identity->hashPubKey,
	   &hexName);
  buildFileName(getHostsDirectory(),
		&hexName,
		fn);
  bzero(&fi.body, sizeof(HELO_Body));
  if (sizeof(HELO_Body) != readFile(fn, 
				    sizeof(HELO_Body), 
				    &fi.body)) {    
    WARNING("WARNING: malformed HELO stored under %s\n",
	  fn);
    return;
  }      
  /* do not forward expired HELOs */
  if (ntohl(fi.body.expirationTime) < now) {
    /* remove HELOs that expired */ 
    unlink(fn);
    delHostFromKnown(identity);
    return;
  }
  /* remove HELOs for blacklisted IPs */ 
  if (YES == isBlacklisted(fi.body.sender.senderIP)) {
    MESSAGE("HELOEXCHANGE: Removing blacklisted file: %s\n",
	    fn);
    unlink(fn); 
    delHostFromKnown(identity);
    return;
  }
  /* everything else: forward with a certain probability */
  ran = randomi(*count);
  if (ran <= log(*count+1)) 
    broadcast(&fi.header, 0);
  xfree(fn,"forwardHelo: filename");
}

/**
 * Forward HELOs from all known hosts to all connected hosts.
 **/
static void forwardHELO(void * unused) {
  int count;

  DEBUG("CRON: enter forwardHELO\n");
  count = forEachHost(NULL, 
		      0, 
		      NULL);
  
  forEachHost((void (*)(HostIdentity*, void *))&forwardHELOHelper,
	      0, /* ignore blacklisting */
	      &count);
  DEBUG("CRON: exit forwardHELO\n");
}

/* end of heloexchange.c */
